/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail;

import java.io.IOException;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.tagext.TagSupport;


/**
 * <p>The main tag class that application will use. This tag will render the current node
 *  to the root node (or vise-versa) using a path separator and specific styles.
 *  The application can setup the separator and the various styles either by specifying
 *   default values in the configuration manager or by specifying attributes on the string.
 *    The object factory will be used to discover the node and the path to the root.
 *    The node discovery is optional. If the node discovery is not specified, it is
 *     assumed a node discovery tag was used on the page to define the current node.
 *      The body of the tag will be used to format the various trails (using variables
 *     defined by this class)..</p>
 * <p>Thread Safety: This class is thread safe in the context of the JSP rendering process.</p>
 *
 *
 * @author mishagam, Khristopher
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 1.0
 */
public class BreadCrumbTrailTag extends TagSupport {
    /** Paths representing trail are shown from left to right, with root  leftmost and current
     *  page rightmost
     * <p>This static constant represents a "pathDirection" constant for formatting the path
     * from the root (on the left) to the current node (on the right).  This variable is only
     * referenced
     * in the doStartTag but is public for a convienence to strategy implementations.</p>
     */
    public static final int ROOT_TO_CURRENT = 0;

    /**
     * Paths representing trail are shown from right to left, with root  rightmost and current
     *  page leftmost.
     * <p>This static constant represents a "pathDirection" constant for formatting the path
     * from the current node (on the left) to the root node (on the right).  This variable is
     * only referenced
     * in the doStartTag but is public for a convienence to strategy implementations.</p>
     */
    public static final int CURRENT_TO_ROOT = 1;

    /**
     * <p>This static constant represents the page scope attribute name that the current node,
     * if found, would be found under.  This constant will be used by a NodeDiscover instance
     * to store the current node in the attribute map and by this class to retrieve..</p>
     */
    public static final String BREADCRUMB_CURRENT_NODE = "breadCrumbCurrentNode";

    /**
     * <p>This static constant represents the nested variable name that the current node to format
     * would be found under.  This constant will be used by this class to define a variable available
     * to a NodeFormatter to format.  Note: this is the page scope attribute name that will be made
     * available as a variable via the BreadCrumbTrailTagExtraInfo class.</p>
     */
    public static final String BREADCRUMB_FORMAT_NODE = "breadCrumbFormatNode";

    /**
     * <p>This static constant represents the nested variable name that the current node's style that
     * will be used to format.  This constant will be used by this class to define a variable
     *  available
     * to a NodeFormatter to format.  Note: this is the page scope attribute name that will be made
     * available as a variable via the BreadCrumbTrailTagExtraInfo class.</p>
     */
    public static final String BREADCRUMB_FORMAT_STYLE_CLASS = "breadCrumbFormatStyleClass";

    /**
     * Direction in which we layout bread crumb trail hierarchy. Possible  values are
     * ROOT_TO_CURRENT and
     * CURRENT_TO_ROOT . setPathDirection  inforces that this variable can have only these
     *  values. ROOT_TO_CURRENT
     * means that root is on the left and current page is one the right.  CURRENT_TO_ROOT
     *  means current page is on the
     * left and root is on the  right. Default is ROOT_TO_CURRENT
     * <p>Represents the direction of the path.  The current allowable values are either
     * ROOT_TO_CURRENT (specifying root [on the left] to current [on the right]) or
     * CURRENT_TO_ROOT (specifying current [on the left] to root [on the right]).
     * This variable is set by the constructor, is mutable and can only be the above values.
     * This variable is referenced by the associated getter/setter methods and the doStartTag
     * methods.</p>
     */
    private int pathDirection = ROOT_TO_CURRENT;

    /**
     * This is separator that is drawn on web page between urls that  form bread crump trail.
     *  Set as property of bread
     * crumb tag.  Default value ;:;
     * <p>Represents the separator that will be drawn between nodes in a path.
     * This variable is set by the constructor, is mutable and will never be null but can be
     * an empty (trim'd) string.
     * This variable is referenced by the associated getter/setter methods and the doAfterBody
     * methods.</p>
     */
    private String pathSeparator = ":";

    /**
     * This is style of separator. This is actual style, not reference, like
     *   color:red font-style:italic.
     * BreadCrumbTag will place all  these styles in the beggining of rendered breadCrumbTrail.
     *     Styles are set as
     * properties of BreadCrumbTag. Default value is null -  means that whatever parent style
     *  present will be used.
     * <p>Represents the separator CSS style that will be used to draw the separator.  This is the
     * actual CSS style that will be rendered to the underlying text.
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd)
     *  string
     * to specify no specific style (ie use the style that is currently in context).
     * This variable is referenced by the associated getter/setter methods and the doAfterBody
     * methods.</p>
     */
    private String pathSeparatorStyle = null;

    /**
     * Style with which Bread Crumb trail links are displayed. Set as  property of BreadCrumbTag.
     *  Default value null
     * and it means that default  links style will be used.
     * <p>Represents the node (non-root and non-current) CSS style that will be used to draw the
     * node's title.  This is the actual CSS style that will be rendered to the underlying link.
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd) string
     * to specify no specific style (ie use the style that is currently in context).
     * This variable is referenced by the associated getter/setter methods and the doAfterBody
     * methods.</p>
     */
    private String nodeStyle = null;

    /**
     * Style with which link of current page in bread crumb trail is displayed.  Set as property
     *  of BreadCrumbTag.
     * Default value null and it means that  style would be the same as in other bread crumb links.
     * <p>Represents the current (only) node CSS style that will be used to draw the
     * node's title.  This is the actual CSS style that will be rendered to the underlying link.
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd)
     *  string
     * to specify no specific style (ie use the style that is currently in context).
     * This variable is referenced by the associated getter/setter methods and the doAfterBody
     * methods.</p>
     */
    private String currentNodeStyle = null;

    /**
     * Style with which link of root page in bread crumb trail is displayed.  Set as property of
     *  BreadCrumbTag. Default
     * value null and it means that  root style would be the same as in other bread crumb links.
     * <p>Represents the root (only) node CSS style that will be used to draw the
     * node's title.  This is the actual CSS style that will be rendered to the underlying link.
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd)
     *  string
     * to specify no specific style (ie use the style that is currently in context).
     * This variable is referenced by the associated getter/setter methods and the doAfterBody
     * methods.</p>
     */
    private String rootNodeStyle = null;

    /** Style of links when mouse moves over them.
     * <p>Represents the mouse over node CSS style that will be used to draw the
     * node's title.  This is the actual CSS style that will be rendered to the underlying link.
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd)
     *  string
     * to specify no specific style (ie use the style that is currently in context).
     * This variable is referenced by the associated getter/setter methods and the doAfterBody
     * methods.</p>
     */
    private String mouseOverStyle = null;

    /**
     * If true - errors in parsing BreadCrumbTrail are catched and  ignored. If false - these
     *  erros are displayed
     * throwing  BreadCrumbExceptions (inherited from JspTagException)
     * <p>Represents whether any exceptions should be silently ignore or not (true to silently ignore,
     * false to let exception pass through to Jsp engine).
     * This variable is set by the constructor and is mutatable.
     * This variable is referenced by the associated getter/setter methods and the doStartTag,
     * doAfterBody and doEndTag methods.</p>
     */
    private boolean silentErrors = false;

    /**
     * <p>Represents the a path override for this specific tag.
     * This variable is set initially to null and is mutatable.  If null, there is no path override
     *  available
     * and normal path discovery should be used.  If non-null, a XmlPathDiscovery class will be
     * constructed with this path and used to return the path specified in this variable.  Please
     *  see
     * XmlPathDiscovery class for details on it's formatting.
     * This variable is referenced by the associated getter/setter methods and the doStartTag
     * methods.</p>
     * <p><b>Implementation notes (version 2.0): </b>new variable.</p>
     */
    private String pathOverride = null;

    /**
     * <p>Represents title that will be used if the node is being discovered by this tag
     *  (ie no other tag has already discovered
     * the node and set the BREADCRUMB_CURRENT_NODE attribute) - otherwise it is ignored.
     * This variable is set initially to null, is mutatable and never will be an empty string
     *  but can be null.  A null indicates that the title
     * should default to the title specified in the PathDiscovery data source.  This variable
     *  is referenced by the
     * associated getter/setter methods and the doStartTag methods.</p>
     * <p><b>Implementation notes (version 2.0): </b>new variable.</p>
     */
    private String title = null;

    /**
     * <p>Represents the iterator over the current node path.  This iterator will initially be
     *  null, be
     * mutatable by the doStartTag and will be non-null during the body processing.  The iterator
     * will be created in the doStartTag from the path discovered by the PathDiscovery
     *  implementation.
     * The node formatter (or body if no node formatter specified) will then be called,
     *  for each node
     * in this itertor.  This variable will only be non-null if a node formatter is not
     *  specified (ie that
     * we need the body to format each node).</p>
     * <p><b>Implementation notes (version 2.0): </b>new variable.</p>
     */
    private ListIterator iterator = null;

    /**
     * <p>Constructs the BreadCrumbTrailTag using the configuration manager to set default values..</p>
     */
    public BreadCrumbTrailTag() {
    }

    /**
     * <p>Returns value of pathDirection property..</p>
     *
     *
     * @return The path direction - either ROOT_TO_CURRENT or CURRENT_TO_ROOT
     * 
     */
    public int getPathDirection() {
        return pathDirection;
    }

    /**
     * <p>Sets value of pathDirection property..</p>
     *
     *
     * @param pathDirection The new path direction
     * @throws IllegalArgumentException if pathDirection is not ROOT_TO_CURRENT or CURRENT_TO_ROOT
     * 
     */
    public void setPathDirection(int pathDirection) {
        if ((pathDirection != ROOT_TO_CURRENT) && (pathDirection != CURRENT_TO_ROOT)) {
            throw new IllegalArgumentException(
                "pathDirection must be equal to BreadCrumbTrailTag.ROOT_TO_CURRENT"
                + " or BreadCrumbTrailTag.CURRENT_TO_ROOT");
        }

        this.pathDirection = pathDirection;
    }

    /**
     * <p>
     * Returns value of pathSeparator property
     * .</p>
     *
     *
     * @return A non-null, possibly empty (trim'd) string representing the path separator
     * 
     */
    public String getPathSeparator() {
        return pathSeparator;
    }

    /**
     * <p>Sets value of pathSeparator property.</p>
     *
     *
     * @param pathSeparator A non-null, possibly empty (trim'd) string representing the
     *  path separator
     * @throws IllegalArgumentException if the pathSeparator is null
     * 
     */
    public void setPathSeparator(String pathSeparator) {
        if (pathSeparator == null) {
            throw new IllegalArgumentException("pathSeparator should not be null");
        }

        this.pathSeparator = pathSeparator;
    }

    /**
     * <p>Returns value of pathSeparatorStyle Property.</p>
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the path separator style
     * 
     */
    public String getPathSeparatorStyle() {
        return pathSeparatorStyle;
    }

    /**
     * <p>Sets value of pathSeparatorStyle property.</p>
     *
     *
     * @param pathSeparatorStyle A possibly null, possibly empty (trim'd) string representing
     *  the path separator style (empty strings will be translated to a null)
     * 
     */
    public void setPathSeparatorStyle(String pathSeparatorStyle) {
        if (pathSeparatorStyle.trim().length() == 0) {
            this.pathSeparatorStyle = null;
        } else {
            this.pathSeparatorStyle = pathSeparatorStyle;
        }
    }

    /**
     * <p>Returns value of nodeStyle property..</p>
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the node style
     * 
     */
    public String getNodeStyle() {
        return nodeStyle;
    }

    /**
     * <p>Sets value of nodeStyle property.</p>
     *
     *
     * @param nodeStyle A possibly null, possibly empty (trim'd) string representing
     *  the node style (an empty string will be translated to a null)
     * 
     */
    public void setNodeStyle(String nodeStyle) {
        if (nodeStyle.trim().length() == 0) {
            this.nodeStyle = null;
        } else {
            this.nodeStyle = nodeStyle;
        }
    }

    /**
     * <p>returns value of currentNodeStyle property.</p>
     *
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the current node style
     * 
     */
    public String getCurrentNodeStyle() {
        return currentNodeStyle;
    }

    /**
     * <p>Sets value of currentNodeStyle property.</p>
     *
     *
     * @param currentNodeStyle A possibly null, possibly empty (trim'd) string representing
     *  the current node style (an empty string will be translated to a null)
     * 
     */
    public void setCurrentNodeStyle(String currentNodeStyle) {
        if (currentNodeStyle.trim().length() == 0) {
            this.currentNodeStyle = null;
        } else {
            this.currentNodeStyle = currentNodeStyle;
        }
    }

    /**
     * <p>returns value of rootNodeStyle property.</p>
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the root node style
     * 
     */
    public String getRootNodeStyle() {
        return rootNodeStyle;
    }

    /**
     * <p>Sets value of rootNodeStyle property.</p>
     *
     *
     * @param rootNodeStyle A possibly null, possibly empty (trim'd) string representing the
     *  root node style (an empty string will be translated to a null)
     * 
     */
    public void setRootNodeStyle(String rootNodeStyle) {
        if (rootNodeStyle.trim().length() == 0) {
            this.rootNodeStyle = null;
        } else {
            this.rootNodeStyle = rootNodeStyle;
        }
    }

    /**
     * <p>returns value of mouseOverStyle property.</p>
     *
     * @return A possibly null, never empty (trim'd) string representing the mouse over style
     * 
     */
    public String getMouseOverStyle() {
        return mouseOverStyle;
    }

    /**
     * <p>Sets value of mouseOverStyle property.</p>
     *
     *
     * @param mouseOverStyle A possibly null, possibly empty (trim'd) string representing
     *  the mouse over style (an empty string will be translated to a null)
     * 
     */
    public void setMouseOverStyle(String mouseOverStyle) {
        if (mouseOverStyle.trim().length() == 0) {
            this.mouseOverStyle = null;
        } else {
            this.mouseOverStyle = mouseOverStyle;
        }
    }

    /**
     * If true - errors in parsing BreadCrumbTrail are catched and  ignored. If false -
     *  these erros are displayed
     * throwing  BreadCrumbExceptions (inherited from JspTagException)
     * <p>Sets value of silentErrors property.</p>
     *
     *
     *
     * @param silentErrors True if exceptions will be silently ignored, false if they
     * should be thrown to the Jsp Engine
     * 
     */
    public void setSilentErrors(boolean silentErrors) {
        this.silentErrors = silentErrors;
    }

    /**
     * If true - errors in parsing BreadCrumbTrail are catched and  ignored. If false -
     * these erros are displayed
     * throwing  BreadCrumbExceptions (inherited from JspTagException)
     * <p>returns value of silentErrors property.</p>
     *
     *
     * @return True if exceptions will be silently ignored, false if they should be thrown
     *  to the Jsp Engine
     * 
     */
    public boolean getSilentErrors() {
        return this.silentErrors;
    }

    /**
     * <p>Prints the style settings out to the specified writer.</p>
     *
     *
     * @return An integer representing what processing to do after this method executes
     * @throws BreadCrumbException if creating of bread crumb trail failed and silentErrors is false
     */
    public int doStartTag() throws BreadCrumbException {
        try {
            /* Try to obtain Map of BreadCrumbNodes */
            HttpServletRequest req = (HttpServletRequest) pageContext.getRequest();
            Map nodes = null;

            try {
                if (this.pageContext != null) {
                    this.iterator = null;

                    //???nodes = parser.parseTrail(this.dataSource, req.getRequestURI());
                } else {
                    //nodes = parseSiteDataSource();
                    if (nodes == null) {
                        /* No data sources are available and silentErrors == true */
                        return super.doStartTag();
                    }
                }
            } catch (Exception bcpe) {
                /* bcpe occurs if an exception occurs while parsing this.dataSource or
                 * this.siteDataSource
                 */
                throwBreadCrumbException("BreadCrumbParserException: " + bcpe.getMessage());

                return super.doStartTag();
            }

            /* Obtained Map of BreadCrumbNodes now do output */
            BreadCrumbNode node = (BreadCrumbNode) nodes.get(req.getRequestURI());

            if (node == null) {
                throwBreadCrumbException("RequestURI key doesn't exist in trail map");

                return super.doStartTag();
            }

            //out = pageContext.getOut();
            //out.print("<span>\n");

            /* output style settings */

            //printStyleSettings();

            /* output bread crumb nodes */
            if (this.pathDirection == CURRENT_TO_ROOT) {
                BreadCrumbNode parentNode = null;

                //???= (BreadCrumbNode) nodes.get(node.getParentUrl());
                if (parentNode == null) {
                    this.iterator = null;

                    /* current node is also root node */

                    //printCurrentRootNode(node.getUrl(), node.getTitle());
                } else {
                    //printCurrentNode(node.getUrl(), node.getTitle());
                    node = parentNode;

                    //???parentNode = (BreadCrumbNode) nodes.get(node.getParentUrl());
                    while (parentNode != null) {
                        /* we are at an intermediate node */

                        //printSeparator();
                        //printNode(node.getUrl(), node.getTitle());
                        node = parentNode;

                        //???parentNode = (BreadCrumbNode) nodes.get(node.getParentUrl());
                    }

                    /* we are at root node */

                    //printSeparator();
                    //printRootNode(node.getUrl(), node.getTitle());
                }
            } else if (this.pathDirection == ROOT_TO_CURRENT) {
                /* first put all BreadCrumbNodes in LinkedList so we can output
                 * from root node to current node
                 */
                List nodesList = new LinkedList();

                while (node != null) {
                    nodesList.add(node);

                    //???node = (BreadCrumbNode) nodes.get(node.getParentUrl());
                }

                /* root node is at end of list so start there and then output
                 * previous nodes
                 */
                ListIterator it = nodesList.listIterator(nodesList.size());
                boolean first = true;

                while (it.hasPrevious()) {
                    node = (BreadCrumbNode) it.previous();

                    if (it.hasPrevious()) {
                        /* we are at an intermediate or root node and there are still
                         * more nodes
                         */
                        if (!first) {
                            this.iterator = null;

                            /* we are at an intermediate node */

                            //printNode(node.getUrl(), node.getTitle());
                        } else {
                            /* we are at the root node and there are still more nodes */

                            //printRootNode(node.getUrl(), node.getTitle());
                            first = false;
                        }

                        //printSeparator();
                    } else {
                        /* we are at current node */
                        if (first) {
                            this.iterator = null;

                            /* current node is also root node */

                            //printCurrentRootNode(node.getUrl(), node.getTitle());
                        } else {
                            this.iterator = null;

                            //printCurrentNode(node.getUrl(), node.getTitle());
                        }
                    }
                }
            }

            //out.print("</span>\n");
            return super.doStartTag();
        } catch (Exception e) {
            throwBreadCrumbException("Exception: " + e.getMessage());
        }

        return 0;
    }

    /**
     * Throws a BreadCrumbException with given message if silentErrors == false.
     *
     * @param message message to include with BreadCrumbException
     *
     * @throws BreadCrumbException if silentErrors == false
     */
    private void throwBreadCrumbException(String message)
        throws BreadCrumbException {
        if (!silentErrors) {
            throw new BreadCrumbException(message);
        }
    }

    /**
     * <p>returns value of pathOverride property.</p>
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the path override
     * 
     */
    public String getPathOverride() {
        return pathOverride;
    }

    /**
     * <p>Sets value of pathOverride property.</p>
     *
     *
     * @param pathOverride A possibly null, possibly empty (trim'd) string representing
     *  the path override (an empty string will be translated to a null)
     * 
     */
    public void setPathOverride(String pathOverride) {
        if (pathOverride.trim().length() == 0) {
            this.pathOverride = null;
        } else {
            this.pathOverride = pathOverride;
        }
    }

    /**
     * <p>Returns value of title property.</p>
     *
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the current node title
     * 
     */
    public String getTitle() {
        return title;
    }

    /**
     * <p>Sets value of title property.</p>
     *
     *
     * @param title A possibly null, possibly empty (trim'd) string representing the title
     *  of the current node (an empty string will be translated to a null)
     * 
     */
    public void setTitle(String title) {
        if (title.trim().length() == 0) {
            this.title = null;
        } else {
            this.title = title;
        }
    }

    /**
     * <p>Get's the next item from the iterator and set's the attribute information for what
     *  needs to be formatted..</p>
     */
    public void doInitBody() {
    }

    /**
     * <p>Get's the body content text and write's it out and starts again.</p>
     *
     *
     * @return Either SKIP_BODY to stop processing the body or EVAL_BODY_AGAIN to process
     *  the next item
     */
    public int doAfterBody() {
        return SKIP_BODY;
    }

    /**
     * <p>Prints the style settings out to the specified writer.</p>
     *
     *
     * @return Always EVAL_PAGE
     * @throws BreadCrumbException if creating of bread crumb trail failed
     *  and silentErrors is false
     */
    public int doEndTag() throws BreadCrumbException {
        return EVAL_PAGE;
    }

    /**
     * <p>Prints the style settings out to the specified writer.</p>
     *
     *
     * @param out The non-null JspWriter to write the style settings to
     * @throws IllegalArgumentException if the out is null
     * @throws IOException if an io occurs writing out the style
     */
    private void printStyleSettings(JspWriter out) throws IOException {
        if (out == null) {
            throw new IllegalArgumentException("out is null.");
        }

        out.print("  <style>\n");

        if (this.pathSeparatorStyle != null) {
            this.iterator = null;

            //printStyle("path-separator-style", this.pathSeparatorStyle);
        }

        if (this.nodeStyle != null) {
            this.iterator = null;

            //printStyle(".node-style a:link", this.nodeStyle);
        }

        if (this.currentNodeStyle != null) {
            this.iterator = null;

            //printStyle(".current-node-style a:link", this.currentNodeStyle);
        }

        if (this.rootNodeStyle != null) {
            this.iterator = null;

            //printStyle(".root-node-style a:link", this.rootNodeStyle);
        }

        if (this.mouseOverStyle != null) {
            //printStyle(".node-style a:hover", this.mouseOverStyle);
            if (this.currentNodeStyle != null) {
                this.iterator = null;

                //printStyle(".current-node-style a:hover", this.mouseOverStyle);
            }

            if (this.rootNodeStyle != null) {
                this.iterator = null;

                //printStyle(".root-node-style a:hover", this.mouseOverStyle);
            }
        }

        out.print("  </style>\n");
    }

    /**
     * <p>Prints the specified style out to the specified writer.</p>
     *
     *
     * @param out The non-null JspWriter to write the style settings to
     * @param style The non-null, non-empty (trim'd) string representing the style to write
     * @param value The non-null, non-empty (trim'd) string representing the value of the style to write
     * @throws IllegalArgumentException if any argument is null or any string argument is an empty (trim'd) string
     * @throws IOException if an io occurs writing out the style
     */
    private void printStyle(JspWriter out, String style, String value)
        throws IOException {
        out.print("    ");
        out.print(style);
        out.print(" {");
        out.print(value);
        out.print("}\n");
    }

    /**
     * <p>Prints the separator (with the separator style) to the specified writer.</p>
     *
     *
     * @param out The non-null JspWriter to write the style settings to
     * @throws IllegalArgumentException if the out is null
     * @throws IOException if an io occurs writing out the style
     */
    private void printSeparator(javax.servlet.jsp.JspWriter out)
        throws IOException {
        out.print("  <span class=path-separator-style>");
        out.print(this.pathSeparator);
        out.print("</span>\n");
    }
}
