/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.nodediscovery;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbTrailTag;

import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.BodyTagSupport;


/**
 * <p>This class provides a tag (only) implementation of node discovery that will allow the user
 * to hardcode the url into the body of this tag and specify the title in the attributes. That
 * hardcoded information is then read by this tag, combined with the title to make a BreadCrumbNode
 *  and that node is put in a page level attribute called BreadCrumbTrailTag.BREADCRUMB_CURRENT_PAGE.</p>
 * <p>Thread Safety: This class is thread safe in the context of the JSP rendering process</p>
 *
 *
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 */
public class InlineTextNodeDiscoveryTag extends BodyTagSupport {
    /**
     * <p>Represents the title associated with this node.  This variable is set initially to null,
     *  is mutable and can be null but not an empty (trim'd) string.
     * A null represents that the title for the current node should come from the PathDiscovery
     *  data source instead.
     * This variable is referenced by the associated getter/setter methods</p>
     */
    private String title = null;

    /**
     * <p>Constructs the InlineTextNodeDiscoveryTag.</p>
     */
    public InlineTextNodeDiscoveryTag() {
        super();
    }

    /**
     * <p>Starts the processing of this tag by telling the support class to evaluate the body.</p>
     *
     *
     * @return Always EVAL_BODY_BUFFERED
     */
    public int doStartTag() {
        return EVAL_BODY_BUFFERED;
    }

    /**
     * <p>Processes the body content. This tag will simply copy the body content, create a
     *  BreadCrumbNode and set it to the attributes as a page level attribute called
     *  BreadCrumbTrailTag.BREADCRUMB_CURRENT_PAGE then clear the body. Please note that
     *  the body is never written out as this tag doesn't render anything</p>
     *
     *
     * @return Always SKIP_BODY
     * @throws BreadCrumbException if the body contained nothing
     */
    public int doAfterBody() throws BreadCrumbException {
        String content = bodyContent.getString();

        if ((content == null) || (content.trim().length() == 0)) {
            throw new BreadCrumbException("body content is null or empty.");
        }

        // create node, set attribute in page context
        BreadCrumbNode node = new BreadCrumbNode(content, getTitle());
        pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_CURRENT_NODE, node, PageContext.PAGE_SCOPE);

        return SKIP_BODY;
    }

    /**
     * <p>Returns value of title property.</p>
     *
     *
     * @return The possibly null, non-empty string representing the title
     */
    public String getTitle() {
        return title;
    }

    /**
     * <p>Sets value of title property.</p>
     *
     *
     * @param title The possibly null, possibly empty string representing the title
     */
    public void setTitle(String title) {
        if (title.trim().length() == 0) {
            this.title = null;
        } else {
            this.title = title;
        }
    }
}
